/**
 *	Copyright (C) 2011-2015 Docuteam GmbH
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 3
 *	as published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package ch.docuteam.converter;

import java.io.*;
import java.lang.management.ManagementFactory;
import java.lang.reflect.InvocationTargetException;
import java.net.ConnectException;
import java.net.SocketException;
import java.util.*;

import javax.swing.SwingWorker;

import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;

import ch.docuteam.converter.exceptions.*;
import ch.docuteam.tools.file.*;
import ch.docuteam.tools.file.exception.*;
import ch.docuteam.tools.id.UniqueID;
import ch.docuteam.tools.os.OperatingSystem;
import ch.docuteam.tools.os.SystemProcess;
import ch.docuteam.tools.out.Logger;
import ch.docuteam.tools.string.DateFormatter;

import com.artofsolving.jodconverter.DocumentConverter;
import com.artofsolving.jodconverter.XmlDocumentFormatRegistry;
import com.artofsolving.jodconverter.openoffice.connection.OpenOfficeConnection;
import com.artofsolving.jodconverter.openoffice.connection.SocketOpenOfficeConnection;
import com.artofsolving.jodconverter.openoffice.converter.OpenOfficeDocumentConverter;
import com.pdftools.converter.Converter;
import com.pdftools.converter.Job;


/**
 * @author denis
 *
 */
public abstract class FileConverter
{
	//	===========================================================================================
	//	========	Structure				=======================================================
	//	===========================================================================================

	//	========	Static Final Public		=======================================================

	//	========	Static Final Private	=======================================================

	static private final String				DefaultMigrationConfigFile = "config/migration-config.xml";
	static private final String				DefaultMigrationHomeFolder = ".";
	static private String					DefaultCommandlineSeparator = "#";
	static private String[] 				AlternativeCommandlineSeparators = new String[] { ";", "&", "@", ",", ":" };

	//	========	Static Public			=======================================================

	//	========	Static Private			=======================================================

	static private String					MigrationConfigFile = DefaultMigrationConfigFile;
	static private String					MigrationHomeFolder = DefaultMigrationHomeFolder;

	static private org.dom4j.Document		MigrationConfigFileDocument;

	static private String					RecentlyUsedFileConverterName = "";

	//	===========================================================================================
	//	========	Main					=======================================================
	//	===========================================================================================

	//	===========================================================================================
	//	========	Methods					=======================================================
	//	===========================================================================================

	//	========	Static Public			=======================================================

	//	--------		Converting			-------------------------------------------------------

	static public File convertFile(File file) throws DocumentException, IllegalArgumentException, SecurityException, IOException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, InterruptedException, IndexOutOfBoundsException, FileIsNotADirectoryException, BadPronomIdException, FileUtilExceptionListException, DROIDCouldNotInitializeException, DROIDNoIdentificationFoundException, DROIDMultipleIdentificationsFoundException, FileConversionException, OOConverterShallNotBeUsedException
	{
		return convertFile(file.getPath());
	}

	static public File convertFile(File sourceFile, File destinationFolder) throws DocumentException, IllegalArgumentException, SecurityException, IOException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, InterruptedException, IndexOutOfBoundsException, FileIsNotADirectoryException, BadPronomIdException, FileUtilExceptionListException, DROIDCouldNotInitializeException, DROIDNoIdentificationFoundException, DROIDMultipleIdentificationsFoundException, FileConversionException, OOConverterShallNotBeUsedException
	{
		return convertFile(sourceFile.getPath(), destinationFolder.getPath());
	}

	static public File convertFile(String fileName) throws DocumentException, IllegalArgumentException, SecurityException, IOException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, InterruptedException, IndexOutOfBoundsException, FileIsNotADirectoryException, BadPronomIdException, FileUtilExceptionListException, DROIDCouldNotInitializeException, DROIDNoIdentificationFoundException, DROIDMultipleIdentificationsFoundException, FileConversionException, OOConverterShallNotBeUsedException
	{
		return convertFile(fileName, null);
	}

	static public File convertFile(String filePath, String destinationFolderPath) throws DocumentException, IllegalArgumentException, SecurityException, IOException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, InterruptedException, IndexOutOfBoundsException, FileIsNotADirectoryException, BadPronomIdException, FileUtilExceptionListException, DROIDCouldNotInitializeException, DROIDNoIdentificationFoundException, DROIDMultipleIdentificationsFoundException, FileConversionException, OOConverterShallNotBeUsedException
	{
		//	Whatever comes in, make the fileNames absolute:
		filePath = new File(filePath).getAbsolutePath();
		if (destinationFolderPath != null)		destinationFolderPath = new File(destinationFolderPath).getAbsolutePath();

		FileWithMetadata file = new FileWithMetadata(filePath);
		if (!file.exists())
		{
			Logger.getLogger().error("File not found: '" + filePath + "'");
			throw new FileNotFoundException(filePath);
		}

		Logger.getLogger().info("Converting file: " + filePath);

		RecentlyUsedFileConverterName = "";

		List<?> migrationInstructions = getMigrationInstructions(file);
		if (migrationInstructions.isEmpty())
		{
			Logger.getLogger().info("... no conversion instruction found.");
			return null;
		}
		if (migrationInstructions.size() >= 2)
		{
			Logger.getLogger().error("Can't handle multiple conversion instructions for file '" + file.getName() + "'");
			throw new IndexOutOfBoundsException("Can't handle multiple conversion instructions for file '" + file.getName() + "'");
		}

		Element migrationInstruction = (Element)migrationInstructions.get(0);
		List<Element> migrationSteps = new ArrayList<Element>();
		List<?> steps = migrationInstruction.selectNodes("step");
		if (steps.isEmpty())
		{
			//	If there is no attribute applicationId, don't convert:
			if (migrationInstruction.attribute("applicationID") == null) return null;
		}
		else
		{
			for (Object step : steps)
			{
				if (step instanceof Element && ((Element)step).attribute("applicationID") != null)
					migrationSteps.add((Element)step);
			}
			if (migrationSteps.isEmpty())		return null;
		}

		//	Note: file.getFormatPronomID() NEVER returns null, migrationInstruction.attributeValue("targetPronom") MIGHT return null:
		if (destinationFolderPath != null)		FileUtil.createFolderMerging(destinationFolderPath);

		Logger.getLogger().info("Treating file as " + migrationInstruction.attributeValue("name") + "...");

		//	This is for the case that the file paths are too long under Windows:
		boolean usingTempFolder = false;
		String originalFilePath = null;
		String originalDestinationFolderPath = null;
		File convertedFile = null;
		try
		{
			if (OperatingSystem.isWindows()
			&&	filePath.length() >= 255 || (destinationFolderPath != null && destinationFolderPath.length() >= 255))
			{
				//	Paths are too long - use temporary folder for conversion:
				usingTempFolder = true;
				originalFilePath = filePath;
				originalDestinationFolderPath = destinationFolderPath;

				filePath = FileUtil.getTempFolder() + "/FileConverter/Temp_" + UniqueID.getString() + "." + FileUtil.asFileNameExtension(filePath);
				FileUtil.copyToOverwriting(originalFilePath, filePath);
				file = new FileWithMetadata(filePath);
				destinationFolderPath = FileUtil.getTempFolder() + "/FileConverter";

				Logger.getLogger().debug("Too long paths, using temp file: " + filePath);
			}

			convertedFile =
				(migrationSteps.isEmpty())
				?	convertFile(file, destinationFolderPath, migrationInstruction)
				:	convertFile(file, destinationFolderPath, migrationSteps);
		}
		catch (InvocationTargetException ex)
		{
			//	If the cause is an OOConverterShallNotBeUsedException, throw it:
			if (ex.getCause().getClass() == OOConverterShallNotBeUsedException.class)		throw (OOConverterShallNotBeUsedException)ex.getCause();

			//	Otherwise rethrow the original exception:
			throw ex;
		}
		finally
		{
			if (usingTempFolder)
			{
				if (convertedFile != null)
				{
					//	Move converted file to original destination:
					String destinationFilePath = originalDestinationFolderPath + "/" + FileUtil.asFileNameWithoutExtension(originalFilePath) + "." + FileUtil.asFileNameExtension(convertedFile.getName());
					//	The returned file is of course the file in the original destination folder:
					convertedFile = FileUtil.moveTo(convertedFile.getPath(), destinationFilePath);
				}

				//	Cleanup temporary files:
				FileUtil.deleteOnExit(filePath);
			}
		}

		return convertedFile;
	}

	//	--------		Accessing			-------------------------------------------------------

	public static String getRecentlyUsedFileConverterName()
	{
		return RecentlyUsedFileConverterName;
	}

	public static void setMigrationConfigFile(String newMigrationConfigFile) throws DocumentException
	{
		MigrationConfigFile = newMigrationConfigFile;
		initialize();
	}

	public static void setMigrationHomeFolder(String newMigrationHomeFolder)
	{
		MigrationHomeFolder = newMigrationHomeFolder;
	}

	//	========	Static Private			=======================================================

	static private File convertFile(FileWithMetadata file, String destinationFolderName, Element migrationInstruction) throws IOException, IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, InterruptedException, FileIsNotADirectoryException, BadPronomIdException, DROIDCouldNotInitializeException, DROIDNoIdentificationFoundException, DROIDMultipleIdentificationsFoundException, FileConversionException, OOConverterShallNotBeUsedException
	{
		if (destinationFolderName == null)
			destinationFolderName = file.getParentFile().getAbsolutePath();
		
		//	this is the filename that the method will return as the result
		String destinationPath = calculateDestinationFilePath(file, destinationFolderName, migrationInstruction.attributeValue("targetExtension"));
		Logger.getLogger().info("Destination file: " + destinationPath);

		//	this is the intermediary filename for the actual conversion that will not pose any problem with special characters that the original filename may contain
		String conversionTargetPath = destinationFolderName + File.separator + UniqueID.getString() + "." + migrationInstruction.attributeValue("targetExtension");
		Logger.getLogger().info("Conversion target file: " + conversionTargetPath);

		String[] commandLine = constructCommandLine(migrationInstruction, new String[]{ file.getAbsolutePath(), conversionTargetPath });
		executeCommandLine(commandLine);

		FileUtil.renameTo(conversionTargetPath, destinationPath);

		String currentPronomId = MetadataProviderDROID.getFileFormatPUID(destinationPath);
		String[] expectedPronomIds = null;
		String expecteds = migrationInstruction.attributeValue("targetPronom");
		if (expecteds != null && !expecteds.isEmpty())
		{
			expectedPronomIds = expecteds.split(",");
			boolean found = false;
			for (String expectedPronomId : expectedPronomIds)
			{
				if (currentPronomId.equalsIgnoreCase(expectedPronomId))
				{
					found = true;
					break;
				}
			}
			if (!found)
			{
				throw new BadPronomIdException(expecteds, currentPronomId);			
			}
		}

		return new File(destinationPath);
	}


	static private File convertFile(FileWithMetadata file, String destinationFolderName, List<?> migrationInstructions) throws IOException, IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, InterruptedException, FileIsNotADirectoryException, BadPronomIdException, FileUtilExceptionListException, DROIDCouldNotInitializeException, DROIDNoIdentificationFoundException, DROIDMultipleIdentificationsFoundException, FileConversionException, OOConverterShallNotBeUsedException
	{
		List<String> tempFilePathsToDelete = new ArrayList<String>();
		String sourcePathString = file.getAbsolutePath();
		String destinationPathString = "";
		int stepNumber = migrationInstructions.size();

		try
		{
			// Loop thru the steps:
			int stepCounter = 1;
			for (Object i: migrationInstructions)
			{
				Element migrationInstruction = (Element)i;

				//	Note: file.getFormatPronomID() NEVER returns null, migrationInstruction.attributeValue("targetPronom") MIGHT return null:
				String[] expectedPronomIds = null;
				String expected = migrationInstruction.attributeValue("targetPronom");

				if (destinationFolderName == null)
					destinationFolderName = file.getParentFile().getAbsolutePath();
				
				//	this is the filename that the method will return as the result
				destinationPathString = calculateDestinationFilePath(file, destinationFolderName, migrationInstruction.attributeValue("targetExtension"));
				Logger.getLogger().info("Step " + stepCounter + "/" + stepNumber + ": Destination file: " + destinationPathString);

				//	Remember this intermediary file for later deletion (excluding the last step):
				if (stepCounter != stepNumber)		tempFilePathsToDelete.add(destinationPathString);

				//	this is the intermediary filename for the actual conversion that will not pose any problem with special characters that the original filename may contain
				String conversionTargetPath = destinationFolderName + File.separator + UniqueID.getString() + "." + migrationInstruction.attributeValue("targetExtension");
				Logger.getLogger().info("Conversion target file: " + conversionTargetPath);

				//	Perform the conversion:
				String[] commandLine = constructCommandLine(migrationInstruction, new String[] { sourcePathString, conversionTargetPath });
				executeCommandLine(commandLine);

				FileUtil.renameTo(conversionTargetPath, destinationPathString);

				String currentPronomId = MetadataProviderDROID.getFileFormatPUID(destinationPathString);
				expectedPronomIds = null;
				if (expected != null && !expected.isEmpty())
				{
					expectedPronomIds = expected.split(",");
					boolean found = false;
					for (String expectedPronomId : expectedPronomIds)
					{
						if (currentPronomId.equalsIgnoreCase(expectedPronomId))
						{
							found = true;
						}
					}
					if (!found)
					{
						throw new BadPronomIdException(expectedPronomIds.toString(), currentPronomId);
					}
				}

				//	Set the input of the next step to be the output of this step:
				sourcePathString = destinationPathString;
				++stepCounter;
			}
		}
		finally
		{
			//	Delete all intermediary files:
			for (String path: tempFilePathsToDelete)		FileUtil.delete(path);
		}

		return new File(destinationPathString);
	}

	//	--------		Initialization		-------------------------------------------------------

	static private void initializeIfNecessary() throws DocumentException
	{
		if (MigrationConfigFileDocument == null)		initialize();
	}


	static private void initialize() throws DocumentException
	{
		Logger.getLogger().info("Initializing File Migration, from " + MigrationConfigFile + "...");

		// Parse MigrationConfig-File:
		MigrationConfigFileDocument = new SAXReader().read(new File(MigrationConfigFile));
	}

	//	--------		Tools				-------------------------------------------------------

	static public String calculateDestinationFilePath(File file, String destinationFolderName, String newExtension)
	{
		String destinationFilePath = destinationFolderName + "/" + FileUtil.asFileNameWithoutExtension(file.getName());

		//	Check if the destination file already exists. If yes, append the current timestamp to the filename:
		if (new File(destinationFilePath + "." + newExtension).exists())
			destinationFilePath += "_" + DateFormatter.getCurrentDateTimeString(DateFormatter.NumericalMSecs);
		destinationFilePath += "." + newExtension;

		return destinationFilePath;
	}


	static public List<?> getMigrationInstructions(FileWithMetadata file) throws DocumentException
	{
		initializeIfNecessary();

		Logger.getLogger().info("Looking for PUID " + file.getFormatPronomID() + "...");
		List<?> instructions = MigrationConfigFileDocument.selectNodes("//puid[@name='" + file.getFormatPronomID() + "']");
		if (!instructions.isEmpty())		return instructions;

		Logger.getLogger().info("Looking for mimetype " + file.getMimeType() + "...");
		instructions = MigrationConfigFileDocument.selectNodes("//mimeType[@name='" + file.getMimeType() + "']");
		if (!instructions.isEmpty())		return instructions;

		String fileNameExtension = file.getName().substring(file.getName().lastIndexOf(".") + 1);
		Logger.getLogger().info("Looking for extension " + fileNameExtension + "...");
		return MigrationConfigFileDocument.selectNodes("//extension[@name='" + fileNameExtension + "']");
	}


	static private String[] constructCommandLine(Element migrationInstruction, String[] args)
	{
		Element applicationElement = (Element) MigrationConfigFileDocument.selectSingleNode("//application[@id='" + migrationInstruction.attributeValue("applicationID") + "']");
		String commandLine = applicationElement.attributeValue("executable");
		RecentlyUsedFileConverterName = applicationElement.attributeValue("name");

		String applicationParameter = applicationElement.attributeValue("parameter");
		String instructionParameter = migrationInstruction.attributeValue("parameter");

		//	instructionParameter take precedence over applicationParameter:
		if (instructionParameter != null && !instructionParameter.trim().isEmpty())
			commandLine += DefaultCommandlineSeparator + instructionParameter;
		else if (applicationParameter != null && !applicationParameter.trim().isEmpty())
			commandLine += DefaultCommandlineSeparator + applicationParameter;

		//Check for unwanted occurences of separator in commandline and replace
		String separator = determineApplicableSeparator(commandLine, args[0]);
		if (!separator.equals(DefaultCommandlineSeparator))
		{
			commandLine = commandLine.replace(DefaultCommandlineSeparator, separator);
		}

		int argIndex = 0;
		while (true)
		{
			String argPlaceholderString = "{[arg" + (argIndex + 1) + "]}";
			if (commandLine.indexOf(argPlaceholderString) == -1)		break;

			commandLine = commandLine.replace(argPlaceholderString, args[argIndex]);
			argIndex = argIndex + 1;
		}

		return commandLine.trim().split(separator);
	}
	
	static private String determineApplicableSeparator(String parameters, String filepath)
	{
		String filename = new File(filepath).getName();
		if (!filename.contains(DefaultCommandlineSeparator))
		{
			return DefaultCommandlineSeparator;
		}
		for (String separator : AlternativeCommandlineSeparators)
		{
			if (!filename.contains(separator) && !parameters.contains(separator))
			{
				return separator;
			}
		}
		return DefaultCommandlineSeparator;
	}


	/**
	 * @throws OOConverterShallNotBeUsedException The call to the converter goes via reflection, that's why this Exception is not visible. It can be thrown by the OOConverter.
	 */
	static private void executeCommandLine(String[] commandLine) throws IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, IOException, InterruptedException, FileIsNotADirectoryException, FileConversionException, OOConverterShallNotBeUsedException
	{
		Logger.getLogger().info("Executing command line: " + Arrays.toString(commandLine));

		if (commandLine[0].startsWith("Class:"))
		{
			//	Application is a Class:
			//	Cut away leading "Class:" from class name:
			String className = commandLine[0].substring(6);

			//	Cut away 1st element of commandLine (= class name):
			commandLine = Arrays.copyOfRange(commandLine, 1, commandLine.length);

			Logger.getLogger().debug("Conversion-process starting");
			Object result = Class.forName(className).getMethod("main", commandLine.getClass()).invoke(null, new Object[] { commandLine });
			Logger.getLogger().info("Conversion-process finished" + ((result != null)? ": " + result: ""));
		}
		else
		{
			int errorCode = SystemProcess.execute(MigrationHomeFolder, commandLine);
			Logger.getLogger().info("Conversion-process finished: " + errorCode);

			if (errorCode != 0)		throw new FileConversionException(errorCode);
		}
	}

	//	===========================================================================================
	//	========	Inner Classes			=======================================================
	//	===========================================================================================

	//	To start the OpenOffice service headless:
	//	Windows:
	//		soffice.exe -headless -nofirststartwizard -accept="socket,host=localhost,port=8100;urp;StarOffice.Service"
	//	OS X:
	//		/Applications/OpenOffice.org.app/Contents/MacOS/soffice.bin -headless -nofirststartwizard -accept="socket,host=localhost,port=8100;urp;StarOffice.Service" &
	/**
	 * This class is a wrapper around the JODConverter class, which in turn is a wrapper around the OpenOffice Document Converter class.
	 * This class is used by the FileConverter reflectively or directly via the main() method, hence it seems to the compiler as if it is unused.
	 * The only method used from the outside is static public void main().
	 */
	static public abstract class OOConverter
	{
		//	Several different default values:
		static private final String				DefaultPathBase = "apps";
		static private final String				DefaultPathMac = DefaultPathBase + "/OpenOffice.org.app";
		static private final String				DefaultPathWindows = DefaultPathBase + "/OpenOffice.org 3";
		static private final String				DefaultPathLinux =  DefaultPathBase + "/openoffice4";
		static private final String				PathRelativeMac = "/Contents/MacOS/soffice.bin";
		static private final String				PathRelativeWindows = "/program/soffice.exe";
		static private final String				PathRelativeLinux = "/program/soffice";

		static private final String				DefaultHost = "localhost";
		static private final int				DefaultPort = 8100;

		static private final String				DefaultJODDocumentFormatsFile = "config/document-formats.xml";

		static private final int				DefaultNumberOfInitializationRetries = 20;


		static private String					Path = OperatingSystem.isWindows()? DefaultPathWindows: (OperatingSystem.isMacOSX()? DefaultPathMac: DefaultPathLinux);
		static private String					Host = DefaultHost;
		static private int						Port = DefaultPort;

		static private String					JODDocumentFormatsFile = DefaultJODDocumentFormatsFile;

		static private boolean					ShallNotBeUsed = false;
		static private boolean					IsInitializing = false;
		static private int						NumberOfInitializationRetries = DefaultNumberOfInitializationRetries;

		static private OpenOfficeConnection		Connection;
		static private DocumentConverter		Converter;
		static private Process					OOProcess;

		//	========	Static Initializer		=======================================================

		//	========	Main					=======================================================

		/**
		 * This main method is the gateway to run the converter. It's NOT just for tests! Don't delete it!
		 * @throws OOConverterShallNotBeUsedException
		 */
		static public void main(String... args) throws IOException, FileUtilExceptionListException, OOConverterShallNotBeUsedException
		{
			if (args.length != 2)
			{
				System.err.println("ERROR: Wrong number of arguments.");
				System.err.println("");
				System.err.println("Usage: FileConverter$OOConverter [path/to/]sourceFile [path/to/]destinationFile");
				return;
			}

			convert2PDF(args[0], args[1]);
		}

		//	========	Static Public			=======================================================

		/**
		 * This must be called BEFORE the OOConverter is initialized!
		 */
		public static void setJODDocumentsFormatFile(String newJODDocumentFormatsFile)
		{
			JODDocumentFormatsFile = newJODDocumentFormatsFile;
		}

		/**
		 * This must be called BEFORE the OOConverter is initialized!
		 */
		public static void setNumberOfInitializationRetries(Integer newNumberOfInitializationRetries)
		{
			if (newNumberOfInitializationRetries == null)		return;
			if (newNumberOfInitializationRetries <= 0)			return;

			NumberOfInitializationRetries = newNumberOfInitializationRetries;
		}

		static public Integer getNumberOfInitializationRetries()
		{
			return NumberOfInitializationRetries;
		}

		static public String getConverterPath()
		{
			return Path;
		}

		static public boolean shallNotBeUsed()
		{
			return ShallNotBeUsed;
		}


		/**
		 * Initialize in a separate thread.
		 */
		static public void initializeDontWait(String ooConverterPath, String ooConverterHost, Integer ooConverterPort)
		{
			if (!(ooConverterHost == null || ooConverterHost.isEmpty()))		Host = ooConverterHost;
			if (!(ooConverterPort == null))										Port = ooConverterPort;
			initializeDontWait(ooConverterPath);
		}


		/**
		 * Initialize in a separate thread.
		 */
		static public void initializeDontWait(String ooConverterPath)
		{
			if (!(ooConverterPath == null || ooConverterPath.isEmpty()))		Path = FileUtil.asCanonicalFileName(ooConverterPath);
			initializeDontWait();
		}


		/**
		 * Initialize in a separate thread.
		 */
		static public void initializeDontWait()
		{
			new SwingWorker<Integer, Object>()
			{
				@Override
				public Integer doInBackground()
				{
					initialize();

					return 0;
				}
			}.execute();
		}


		static public void initialize(String ooConverterPath, String ooConverterHost, Integer ooConverterPort)
		{
			if (!(ooConverterHost == null || ooConverterHost.isEmpty()))		Host = ooConverterHost;
			if (!(ooConverterPort == null))										Port = ooConverterPort;
			initialize(ooConverterPath);
		}


		static public void initialize(String ooConverterPath)
		{
			if (!(ooConverterPath == null || ooConverterPath.isEmpty()))		Path = FileUtil.asCanonicalFileName(ooConverterPath);
			initialize();
		}


		static public void initialize()
		{
			//	Maybe the Converter exists already?
			if (Converter != null)		return;

			//	Converter does not exist. But maybe the OO service is just being initialized in another thread? If yes, wait until initialization is finished:
			while(IsInitializing)
			{
				Logger.getLogger().debug("Waiting for initialization to finish...");
				try { Thread.sleep(500); } catch (InterruptedException x){}
			}

			//	Maybe the Converter exists now?
			if (Converter != null)		return;

			//	Try to create the converter - if it succeeded, the OO Service was already running:
			Converter = createConverter();
			if (Converter != null)		return;

			String ooConverterCompletePath = Path;
			if (OperatingSystem.isWindows())		ooConverterCompletePath += PathRelativeWindows;
			else if (OperatingSystem.isMacOSX())	ooConverterCompletePath += PathRelativeMac;
			else if (OperatingSystem.isLinux())		ooConverterCompletePath += PathRelativeLinux;

			Logger.getLogger().debug("Starting OO Service on host:'" + Host + "' port:'" + Port + "' path:'" + ooConverterCompletePath + "'");

			//	The OO Service is not running, so start it:
			try
			{
				//	This is when initializing in an own thread:
				IsInitializing = true;

				try
				{
					//	Is the OO application where it's supposed to be?
					if (!new File(ooConverterCompletePath).exists())
					{
						if (Path.startsWith(DefaultPathBase))
						{
							Logger.getLogger().info("Could not find the default OO application for converting files in folder: " + Path);

							//	The OO path is the defaultPath but OO is not there - so I suppose that OO shall not be used at all:
							ShallNotBeUsed = true;
							return;
						}

						System.out.println("I was told that the OO application is right there.");
						System.out.println("Maybe this location is wrong or the OO application is not installed there?");

						throw new FileNotFoundException(ooConverterCompletePath);
					}

					//	Start the OOProcess:
					OOProcess = Runtime.getRuntime().exec(new String[]{ooConverterCompletePath, "-headless", "-nofirststartwizard", "-accept=socket,host=" + Host + ",port=" + Port + ";urp;StarOffice.Service"});
				}
				catch (IOException x)
				{
					x.printStackTrace();
					return;
				}

				//	Now that the OOProcess is running, create the Converter, retry several times, wait 1 sec between each try:
				for (int i = NumberOfInitializationRetries; i > 0; i--)
				{
					Logger.getLogger().debug(i + ":");

					Converter = createConverter();
					if (Converter != null)
					{
						Logger.getLogger().debug("OOConverter established");
						break;
					}

					//	Wait a sec and then retry:
					try { Thread.sleep(1000); } catch (InterruptedException x){}
				}
			}
			finally
			{
				IsInitializing = false;
			}
		}


		static public void disconnect()
		{
			if (Connection != null)
			{
				Logger.getLogger().debug("Disconnecting OOConverter...");

				//	Sometimes, connection.disconnect() throws a NullPointerException for no obvious reason:
				try
				{
					Connection.disconnect();
				}
				catch(NullPointerException x) {}
				finally
				{
					Connection = null;
				}
			}

			if (OOProcess != null)
			{
				Logger.getLogger().debug("Destroying OOProcess...");

				OOProcess.destroy();

				//	Wait a sec (otherwise it won't work):
				try { Thread.sleep(1000); } catch (InterruptedException x){}
			}

			//	Clean-up thoroughly:
			Path = OperatingSystem.isWindows()? DefaultPathWindows: (OperatingSystem.isMacOSX()? DefaultPathMac: DefaultPathLinux);
			Host = DefaultHost;
			Port = DefaultPort;

			JODDocumentFormatsFile = DefaultJODDocumentFormatsFile;

			NumberOfInitializationRetries = DefaultNumberOfInitializationRetries;

			Converter = null;
			ShallNotBeUsed = false;
		}


		static public void convert2PDF(String sourceFileName, String destinationFileName) throws IOException, FileUtilExceptionListException, OOConverterShallNotBeUsedException
		{
			initializeIfNecessary();
			if (ShallNotBeUsed)		throw new OOConverterShallNotBeUsedException();

			Logger.getLogger().info("Converting file: '" + sourceFileName + "' to PDF: '" + destinationFileName + "'");

			File sourceFile = new File(sourceFileName);
			File destinationFile = new File(destinationFileName);

			if (!sourceFile.exists())		throw new FileNotFoundException(sourceFileName);
			FileUtil.createFolderMerging(destinationFile.getParentFile());
			FileUtil.delete(destinationFile);

			try
			{
				Converter.convert(sourceFile, destinationFile);
			}
			catch (NullPointerException x)
			{
				Logger.getLogger().error("File conversion failed - maybe the OpenOffice service is not running?");
				throw x;
			}
		}


		static public boolean isInstalledLocallyForWindows()
		{
			return new File(DefaultPathWindows).exists();
		}

		static public boolean isInstalledLocallyForOSX()
		{
			return new File(DefaultPathMac).exists();
		}

		static public boolean isInstalledLocallyForLinux()
		{
			return new File(DefaultPathLinux).exists();
		}

		/**
		 * Return the remote path for Windows if the current OS is Windows and the path is not the local default path.
		 * Otherwise return an empty string.
		 * @return
		 */
		static public String getRemotePathForWindows()
		{
			return OperatingSystem.isWindows() && !Path.equals(DefaultPathWindows)? Path: "";
		}

		/**
		 * Return the remote path for OSX if the current OS is OSX and the path is not the local default path.
		 * Otherwise return an empty string.
		 * @return
		 */
		static public String getRemotePathForOSX()
		{
			return OperatingSystem.isMacOSX() && !Path.equals(DefaultPathMac)? Path: "";
		}

		/**
		 * Return the remote path for Linux if the current OS is Linux and the path is not the local default path.
		 * Otherwise return an empty string.
		 * @return
		 */
		static public String getRemotePathForLinux()
		{
			return OperatingSystem.isLinux() && !Path.equals(DefaultPathLinux)? Path: "";
		}

		//	========	Static Private			=======================================================

		static private void initializeIfNecessary()
		{
			if (Converter == null)		initialize();
		}


		static private DocumentConverter createConverter()
		{
			DocumentConverter converter = null;

			try
			{
				Logger.getLogger().debug("Connecting to OO Service...");

				(Connection = new SocketOpenOfficeConnection(Port)).connect();
				try
				{
					Logger.getLogger().debug("Trying to initialize OO Service with initialization file '" + JODDocumentFormatsFile + "'");
					converter = new OpenOfficeDocumentConverter(Connection, new XmlDocumentFormatRegistry(new FileInputStream(JODDocumentFormatsFile)));
				}
				catch (FileNotFoundException e)
				{
					Logger.getLogger().debug("Initialization file '" + JODDocumentFormatsFile + "' not found, using defaults");
					converter = new OpenOfficeDocumentConverter(Connection);
				}

				Logger.getLogger().debug("... OO Service connected");
			}
			catch (ConnectException e)
			{
				Logger.getLogger().debug(e.toString());
			}

			return converter;
		}
	}




	/**
	 * This class is a dummy conversion class that simply creates a copy of a file.
	 * This class is used by the FileConverter reflectively, hence it seems to the compiler as if it is unused.
	 * The only method used from the outside is static public void main().
	 */
	@SuppressWarnings("unused")
	static private class Copy
	{
		//	========	Static Public			=======================================================

		/**
		 * This main method is the gateway to run the converter. It's NOT just for tests! Don't delete it!
		 */
		static public void main(String... args) throws IOException, FileUtilExceptionListException
		{
			if (args.length != 2)
			{
				System.err.println("ERROR: Wrong number of arguments.");
				System.err.println("");
				System.err.println("Usage: FileConverter$OOConverter [path/to/]sourceFile [path/to/]destinationFile");
				return;
			}

			FileUtil.copyToOverwriting(args[0], args[1]);
		}
	}


	/**
	 * This class is a wrapper around the PDFTools Converter.
	 * NOTE: PDFTools Converter works only under Windows, NOT under OSX or Linux!
	 * NOTE: PDFTools requires the following preconditions to run correctly:
	 * 		1:	The dll files "O2PProxyAPI.dll" and "O2PProxyNET.dll" (in the correct 32Bit/64Bit version for the client computer) have to be present.
	 * 		2:	The dll "O2PProxyNET.dll" has to be registered in the "Global Assembly Cache". Refer to: <http://msdn.microsoft.com/en-us/library/ex0ss12c%28v=vs.110%29.aspx>.
	 * 		3:	The Java application has to be started with the VM Parameter: -Djava.library.path=[dllPath],
	 * 			where [dllPath] is the relative or absulute path to the folder containing O2PProxyAPI.dll (without trailing slash).
	 * 		4:	If not initialized explicitly with an URL pointing to the (remote) PDFTools service, assume a local installation.
	 * This class is used by the FileConverter reflectively or directly via the main() method, hence it seems to the compiler as if it is unused.
	 * The only method used from the outside is static public void main().
	 */
	static public abstract class PDFToolsConverter
	{
		//	===========================================================================================
		//	========	Structure				=======================================================
		//	===========================================================================================

		//	========	Static Final Public		=======================================================

		//	========	Static Final Private	=======================================================

		//	========	Static Public			=======================================================

		//	========	Static Private			=======================================================

		static private Converter			ConverterInstance;
		static private Job					JobInstance;

		//	===========================================================================================
		//	========	Main					=======================================================
		//	===========================================================================================

		/**
		 * This main method is the gateway to run the converter. It's NOT just for tests! Don't delete it!
		 * @throws SocketException
		 * @throws PDFToolsConverterBadVMArgumentException
		 */
		static public void main(String... args) throws SocketException, PDFToolsConverterBadVMArgumentException
		{
			if (args.length != 2 && args.length != 3)
			{
				System.err.println("ERROR: Wrong number of arguments.");
				System.err.println("");
				System.err.println("Usage: FileConverter$PDFToolsConverter [path/to/]sourceFile [path/to/]destinationFile [options]");
				return;
			}

			if (args.length == 2)			convert2PDF(args[0], args[1]);
			else if (args.length == 3)		convert2PDF(args[0], args[1], args[2]);
		}

		//	===========================================================================================
		//	========	Methods					=======================================================
		//	===========================================================================================

		//	========	Static Initializer		=======================================================

		//	========	Static Public			=======================================================

		//	--------		Initializing		-------------------------------------------------------

		/**
		 * Initialize local PDF-Tools converter.
		 * @throws SocketException
		 * @throws PDFToolsConverterBadVMArgumentException
		 */
		static public void initialize() throws SocketException, PDFToolsConverterBadVMArgumentException
		{
			initialize(null);
		}


		/**
		 * Initialize remote PDF-Tools converter.
		 * @param url The service URL for connecting to the document converter service. When null is specified, the default "tcp://localhost:7981/O2PService" is used.
		 * @throws SocketException
		 * @throws PDFToolsConverterBadVMArgumentException
		 */
		static public void initialize(String url) throws SocketException, PDFToolsConverterBadVMArgumentException
		{
			checkRequiredVMArgument();

			ConverterInstance = Converter.createConverter(url);
			JobInstance = ConverterInstance.createJob();
		}


		/**
		 * Disconnect PDF-Tools converter and release all resources.
		 */
		static public void disconnect()
		{
			if (JobInstance != null)		JobInstance.destroyObject();
			if (ConverterInstance != null)	ConverterInstance.destroyObject();
		}

		//	--------		Accessing			-------------------------------------------------------
		//	--------		Inquiring			-------------------------------------------------------
		//	--------		Business Ops		-------------------------------------------------------

		static public void convert2PDF(String sourceFilePath, String destinationFilePath) throws SocketException, PDFToolsConverterBadVMArgumentException
		{
			convert2PDF(sourceFilePath, destinationFilePath, "");
		}


		static public void convert2PDF(String sourceFilePath, String destinationFilePath, String options) throws SocketException, PDFToolsConverterBadVMArgumentException
		{
			initializeIfNecessary();
			if (JobInstance == null)
			{
				Logger.getLogger().error("PDF-Tools Converter can not be initialized");
				return;
			}

			Logger.getLogger().info("Converting: '" + sourceFilePath + "' to: '" + destinationFilePath + "'");

			boolean ok;
			try
			{
				ok = JobInstance.createOutput(destinationFilePath);
				if (!ok)
				{
					Logger.getLogger().error(JobInstance.getLastError().Code);
					Logger.getLogger().error(JobInstance.getLastError().Text);
					return;
				}

				ok = JobInstance.appendDoc(sourceFilePath, options);
				if (!ok)
				{
					Logger.getLogger().error(JobInstance.getLastError().Code);
					Logger.getLogger().error(JobInstance.getLastError().Text);
				}
			}
			finally
			{
				ok = JobInstance.close();
				if (!ok)
				{
					Logger.getLogger().error(JobInstance.getLastError().Code);
					Logger.getLogger().error(JobInstance.getLastError().Text);
				}
			}
		}

		//	--------		Persistence			-------------------------------------------------------
		//	--------		Support				-------------------------------------------------------
		//	--------		Utilities			-------------------------------------------------------
		//	---------		Misc				-------------------------------------------------------
		//	--------		Debugging			-------------------------------------------------------
		//	---------		Temporary			-------------------------------------------------------

		//	========	Static Private			=======================================================

		//	--------		Initializing		-------------------------------------------------------

		static private void initializeIfNecessary() throws SocketException, PDFToolsConverterBadVMArgumentException
		{
			if (JobInstance == null)		initialize();
		}


		static private void checkRequiredVMArgument() throws PDFToolsConverterBadVMArgumentException
		{
			String javaLibraryPath = null;
			for (String vmArg: ManagementFactory.getRuntimeMXBean().getInputArguments())
			{
				if (vmArg.startsWith("-Djava.library.path"))
				{
					javaLibraryPath = vmArg.substring(vmArg.lastIndexOf("=") + 1).trim();
					break;
				}
			}

			if (javaLibraryPath == null)
				throw new PDFToolsConverterBadVMArgumentException("The Java VM Argument '-Djava.library.path' is not defined. It has to point to the folder containing the file 'O2PProxyAPI.dll'");

			if (!new File(javaLibraryPath + "/O2PProxyAPI.dll").exists())
				throw new PDFToolsConverterBadVMArgumentException("The Java VM Argument '-Djava.library.path' doesn't point to the folder containing the file 'O2PProxyAPI.dll'");
		}

		//	--------		Accessing			-------------------------------------------------------
		//	--------		Inquiring			-------------------------------------------------------
		//	--------		Business Ops		-------------------------------------------------------
		//	--------		Persistence			-------------------------------------------------------
		//	--------		Support				-------------------------------------------------------
		//	--------		Utilities			-------------------------------------------------------
		//	---------		Misc				-------------------------------------------------------
		//	--------		Debugging			-------------------------------------------------------
		//	---------		Temporary			-------------------------------------------------------

		//	===========================================================================================
		//	========	Inner Classes			=======================================================
		//	===========================================================================================

	}

}
