/**
 *	Copyright (C) 2011-2015 Docuteam GmbH
 *
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 3
 *	as published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package ch.docuteam.tools.exception;
/**
 *
 */


import java.util.Date;

import ch.docuteam.tools.string.DateFormatter;

/**
 * An instance of this class represents an exception. These exceptions are not thrown but are collected in the class <a href="./ExceptionCollector.html">ExceptionCollector</a>.<p>
 * The public interface is kept as tiny as possible. When an instance of Exception is created, it is automatically added to the ExceptionCollector.
 * @author denis
 *
 */
public class Exception
{
	//	===========================================================================================
	//	========	Structure				=======================================================
	//	===========================================================================================

	//	========	Static Final Public		=======================================================

	//	========	Static Final Private	=======================================================

	/**
	 * The Date format used in toString()
	 */
	static final private String			DateFormatString = "HH:mm:ss.SSS";

	//	========	Static Public			=======================================================

	//	========	Static Private			=======================================================

	//	========	Instance Public			=======================================================

	//	========	Instance Private		=======================================================

	/**
	 * The timestamp when this exception occurred.
	 */
	private Date			timeStamp = new Date();

	/**
	 * The class name where this exception occurred.
	 */
	private String			className;

	/**
	 * The method name where this exception occurred.
	 */
	private String			methodName;

	/**
	 * The file name in whose class this exception occurred.
	 */
	private String			fileName;

	/**
	 * The line number in the file where this exception occurred.
	 */
	private int				lineNumber;

	/**
	 * The message of this exception. When an exception is created out of a Throwable, message is set to this Throwable's message.
	 */
	private String			message;

	/**
	 * The original exception (optional, might be null).
	 */
	private Throwable		exception;				//	Optional, might be null

	//	===========================================================================================
	//	========	Methods					=======================================================
	//	===========================================================================================

	//	========	Static Initializer		=======================================================

	//	========	Constructors Public		=======================================================

	//	========	Constructors Private	=======================================================

	/**
	 * Constructor without an original exception but only with a message string.
	 * In this case, the class name and method name are taken from the place where this instance was created.
	 */
	protected Exception(String message)
	{
		this.message = message;

		StackTraceElement caller = Thread.currentThread().getStackTrace()[3];
		this.className = caller.getClassName();
		this.methodName = caller.getMethodName();
		this.lineNumber = caller.getLineNumber();
		this.fileName = caller.getFileName();
	}


	/**
	 * Constructor with an original exception.
	 * In this case, the class name and method name are taken from the place where this original exception was thrown.
	 * The message is taken from the original exception.
	 */
	protected Exception(Throwable x)
	{
		this.exception = x;
		this.message = x.getMessage();

		StackTraceElement caller = x.getStackTrace()[0];
		this.className = caller.getClassName();
		this.methodName = caller.getMethodName();
		this.lineNumber = caller.getLineNumber();
		this.fileName = caller.getFileName();
	}


	//	========	Static Public			=======================================================
	//				Factories:

	/**
	 * Create a new instance of Exception and add it to the ExceptionCollector.
	 */
	static public void remember(String message)
	{
		ExceptionCollector.add(new Exception(message));
	}


	/**
	 * Create a new instance of Exception and add it to the ExceptionCollector.
	 */
	static public void remember(Throwable x)
	{
		ExceptionCollector.add(new Exception(x));
	}

	//	========	Static Private			=======================================================

	//	========	Instance Public			=======================================================

	//	--------		Accessing			-------------------------------------------------------

	public Date getTimeStamp()
	{
		return this.timeStamp;
	}


	public String getClassName()
	{
		return this.className;
	}


	public String getMethodName()
	{
		return this.methodName;
	}


	public String getFileName()
	{
		return this.fileName;
	}


	public int getLineNumber()
	{
		return this.lineNumber;
	}


	public String getMessage()
	{
		return this.message;
	}


	public Throwable getException()
	{
		return this.exception;
	}


	/**
	 * The description is either the message, if no exception is present, or the exception's toString.
	 * @return
	 */
	public String getDescription()
	{
		return (this.exception == null)? this.message: this.exception.toString();
	}


	public void printStackTrace()
	{
		if (this.exception != null)		this.exception.printStackTrace();
		else
		{
			System.err.print(this.getClassName());
			System.err.print(": ");
			System.err.print(this.message);
			System.err.println();
			System.err.print("\tat ");
			System.err.print(this.className);
			System.err.print(".");
			System.err.print(this.methodName);
			System.err.print("(");
			System.err.print(this.fileName);
			System.err.print(":");
			System.err.print(this.lineNumber);
			System.err.print(")");
			System.err.println();
		}
	}

	//	--------		Inquiring			-------------------------------------------------------
	//	--------		Interface			-------------------------------------------------------
	//	--------		Business Ops		-------------------------------------------------------
	//	--------		Persistence			-------------------------------------------------------
	//	--------		Support				-------------------------------------------------------
	//	--------		Utilities			-------------------------------------------------------
	//	--------		Debugging			-------------------------------------------------------

	@Override
	public String toString()
	{
		return new StringBuilder("Exception: ")
			.append(this.className)
			.append(".")
			.append(this.methodName)
			.append("()[")
			.append(this.lineNumber)
			.append("]: ")
			.append(this.getDescription())
			.toString();
	}


	public String toStringLong()
	{
		return new StringBuilder("[Exception: ")
			.append(DateFormatter.getDateTimeString(this.timeStamp, DateFormatString))
			.append(":")
			.append(this.className)
			.append(".")
			.append(this.methodName)
			.append("(): ")
			.append(this.getDescription())
			.append("]")
			.toString();
	}

	//	---------		Temporary			-------------------------------------------------------

	//	========	Instance Private		=======================================================

	//	--------		Initializing		-------------------------------------------------------
	//	--------		Accessing			-------------------------------------------------------
	//	--------		Inquiring			-------------------------------------------------------
	//	--------		Business Ops		-------------------------------------------------------
	//	--------		Persistence			-------------------------------------------------------
	//	--------		Support				-------------------------------------------------------
	//	--------		Utilities			-------------------------------------------------------
	//	--------		Debugging			-------------------------------------------------------
	//	---------		Temporary			-------------------------------------------------------

}
